#include "ggap/help-p.h"
#include "ggap/dialogs.h"
#include "ggap/gap.h"
#include "ggap/prefs.h"
#include "ggap/prefsdialog.h"
#include <moo/debug.h>
#include <QtGui>

using namespace ggap;

QPointer<HelpBrowser> HelpBrowserPrivate::instance;

HelpView::HelpView(QWidget *parent) :
    QTextBrowser(parent),
    impl(this)
{
    applyPrefs();
}

HelpView::~HelpView()
{
}

HelpViewPrivate::HelpViewPrivate(HelpView *p) :
    pub(p)
{
}

void HelpView::setFontFromPrefs()
{
    setFont(prefsValue(Prefs::HelpFont));
}

void HelpView::applyPrefs()
{
    setFontFromPrefs();
}


QString HelpBrowser::parseUrl(const QString &str)
{
    if (!str.startsWith("gfile:"))
        return str;

    QString path = str.mid(6);
    QUrl url = QUrl::fromLocalFile(path);
    url.setScheme("gfile");
    return url.toString();
}

void HelpView::setSource(const QUrl &url)
{
    QUrl realUrl = url;
    if (realUrl.scheme() == "gfile")
        realUrl.setScheme("file");
    else if (realUrl.scheme() == "help")
        realUrl.setScheme("qrc");

    QTextBrowser::setSource(realUrl);
    QString p = source().toLocalFile();
    emit upAvailable(!p.isEmpty() && !QFileInfo(p).isRoot());
}

void HelpView::up()
{
    QUrl url = source();
    QString path = url.toLocalFile();
    if (!path.isEmpty())
        setSource(QUrl::fromLocalFile(QFileInfo(path).path()));
}

void HelpView::home()
{
    QString path = prefsValue(Prefs::HelpHomePage);
    if (path.isEmpty() || !QFileInfo(path).exists())
        path = GapOptions().docDir() + "/htm/ref/chapters.htm";
    if (QFileInfo(path).exists())
        setSource(QUrl::fromLocalFile(path));
    else
        setSource(QUrl());
}

static QVariant loadDir(const QFileInfo &fi)
{
    QString str;
    QTextStream s(&str, QIODevice::WriteOnly);

    s << "<HTML>\n";
    s << " <HEAD>\n";
    s << "  <TITLE>" << Qt::escape(QDir::toNativeSeparators(fi.filePath())) << "</TITLE>\n";
    s << " </HEAD>\n";
    s << " <BODY>\n";
    s << "<H1>" << Qt::escape(QDir::toNativeSeparators(fi.filePath())) << "</H1>\n";
    s << "<HR><PRE>\n";

    if (!fi.isRoot())
        s << "<IMG SRC=\":/icons16/go-up.png\" ALT=\"[UP]\"> <A HREF=\""
          << QUrl::fromLocalFile(fi.path()).toString() << "\">..</A>\n";

    QString dirsep = QDir::toNativeSeparators("/");
    QDir d(fi.filePath());
    const QFileInfoList list = d.entryInfoList(QDir::AllEntries | QDir::NoDotAndDotDot,
                                               QDir::Name | QDir::IgnoreCase);
    foreach (const QFileInfo &entry, list)
        if (entry.isDir())
            s << "<IMG SRC=\":/icons16/folder.png\" ALT=\"[DIR]\"> <A HREF=\""
              << QUrl::fromLocalFile(entry.filePath()).toString()
              << "\">" << Qt::escape(entry.fileName()) << dirsep << "</A>\n";
        else
            s << "<IMG SRC=\":/icons16/file.png\" ALT=\"[FILE]\"> <A HREF=\""
              << QUrl::fromLocalFile(entry.filePath()).toString()
              << "\">" << Qt::escape(entry.fileName()) << "</A>\n";

    s << "</PRE><HR>\n";
    s << "</BODY></HTML>\n";

    return str;
}

static inline QString htmlFromPlainText(const QString &text)
{
    QString html = "<html><body><pre>";
    html.append(Qt::escape(text));
    html.append("</pre></body></html>");
    return html;
    return Qt::convertFromPlainText(text);
}

static inline bool isHtml(const QString &s)
{
    return Qt::mightBeRichText(s);
}

static inline void fixHtml(QString &s)
{
    static QRegExp ffffff("<(body bgcolor|BODY BGCOLOR)=\"(ffffff|FFFFFF)\">");
    s.replace(ffffff, "<body bgcolor=\"#ffffff\">");
}

static inline QString checkLoadHtml(QString &s)
{
    if (isHtml(s))
    {
        fixHtml(s);
        return s;
    }
    else
    {
        return htmlFromPlainText(s);
    }
}

QVariant HelpView::checkLoadHtml(const QVariant &data, const QUrl &)
{
    if (data.isNull())
        return QString("<html><body><h1>%1</h1></body></html>").arg(tr("Not found"));

    if (data.type() == QVariant::String)
    {
        QString s = data.toString();
        return ::checkLoadHtml(s);
    }
    else if (data.type() == QVariant::ByteArray)
    {
        const QByteArray ba = data.toByteArray();
        QTextCodec *codec = Qt::codecForHtml(ba);
        QString s = codec->toUnicode(ba);
        return ::checkLoadHtml(s);
    }
    else
    {
        return data;
    }
}

QVariant HelpView::checkLoadCss(const QVariant &data, const QUrl &url)
{
    if (!data.isNull())
        return data;

    QString filename = url.toLocalFile();
    QRegExp re("(.*/pkg/[^/]+/doc/)manual.css");
    re.setCaseSensitivity(Qt::CaseInsensitive);

    if (re.exactMatch(filename))
    {
        QString newFilename = re.cap(1) + "gapdoc.css";
        if (QFileInfo(newFilename).exists())
        {
            mDebug("misc") << filename << "not found, trying" << newFilename;
            return loadResource(QTextDocument::StyleSheetResource,
                                QUrl::fromLocalFile(newFilename));
        }
    }

    return data;
}

QVariant HelpView::loadResource(int type, const QUrl &url)
{
    if (!url.isRelative())
    {
        QFileInfo fi(url.toLocalFile());
        if (fi.isDir())
            return loadDir(fi);
    }

    const QVariant data = QTextBrowser::loadResource(type, url);

    switch (type)
    {
        case QTextDocument::StyleSheetResource:
            return checkLoadCss(data, url);
        case QTextDocument::HtmlResource:
            return checkLoadHtml(data, url);
        default:
            return data;
    }
}


HelpBrowser::HelpBrowser() :
    moo::AppWindow("help"),
    impl(this)
{
    impl->setupUi();
}

HelpBrowser::~HelpBrowser()
{
}

void HelpBrowser::showWindow(const QUrl &url)
{
    if (!HelpBrowserPrivate::instance)
    {
        HelpBrowserPrivate::instance = new HelpBrowser;
        HelpBrowserPrivate::instance->show();
    }

    HelpBrowser *b = HelpBrowserPrivate::instance;
    HelpView *v = b->view();
    b->raise();
    b->activateWindow();

    if (v->source().isEmpty() && !url.isValid())
        v->home();
    else if (!url.isEmpty())
        v->setSource(url);
}

void HelpBrowser::printDocument(QPrinter *printer)
{
    view()->print(printer);
}

void HelpBrowser::setupPrintDialog(QPrintDialog *dlg)
{
    if (view()->textCursor().hasSelection())
        dlg->addEnabledOption(QAbstractPrintDialog::PrintSelection);
}

HelpBrowserPrivate::HelpBrowserPrivate(HelpBrowser *p) :
    pub(p), addressEntry(0)
{
}

void HelpBrowserPrivate::setupUi()
{
    M_Q(HelpBrowser);

    ui.setupUi(q);

    QMenu *bookmarkMenu = createBookmarkMenu(ui.actionAddBookmark, ui.menubar);
    ui.menubar->insertMenu(ui.menuHelp->menuAction(), bookmarkMenu);
    QObject::connect(bookmarkMenu, SIGNAL(itemActivated(int)), ui.bookmarkList, SLOT(activateBookmark(int)));

    QObject::connect(ui.helpView, SIGNAL(highlighted(QString)), ui.statusbar, SLOT(showMessage(QString)));
    QObject::connect(ui.helpView, SIGNAL(forwardAvailable(bool)), ui.actionGoForward, SLOT(setEnabled(bool)));
    QObject::connect(ui.helpView, SIGNAL(backwardAvailable(bool)), ui.actionGoBackward, SLOT(setEnabled(bool)));
    QObject::connect(ui.helpView, SIGNAL(upAvailable(bool)), ui.actionGoUp, SLOT(setEnabled(bool)));
    QObject::connect(ui.helpView, SIGNAL(copyAvailable(bool)), ui.actionCopy, SLOT(setEnabled(bool)));
    QObject::connect(ui.actionGoForward, SIGNAL(triggered()), ui.helpView, SLOT(forward()));
    QObject::connect(ui.actionGoBackward, SIGNAL(triggered()), ui.helpView, SLOT(backward()));
    QObject::connect(ui.actionGoHome, SIGNAL(triggered()), ui.helpView, SLOT(home()));
    QObject::connect(ui.actionCopy, SIGNAL(triggered()), ui.helpView, SLOT(copy()));
    QObject::connect(ui.actionGoUp, SIGNAL(triggered()), ui.helpView, SLOT(up()));
    QObject::connect(ui.bookmarkDelete, SIGNAL(clicked()), ui.bookmarkList, SLOT(deleteSelected()));
    QObject::connect(ui.bookmarkAdd, SIGNAL(clicked()), ui.actionAddBookmark, SLOT(trigger()));
    QObject::connect(ui.bookmarkList, SIGNAL(urlActivated(QUrl)), ui.helpView, SLOT(setSource(QUrl)));

    q->connectAction(ui.actionFind, ui.searchBox, SLOT(find()));
    q->connectAction(ui.actionFindNext, ui.searchBox, SLOT(findNext()));
    q->connectAction(ui.actionFindPrevious, ui.searchBox, SLOT(findPrevious()));
    ui.searchBox->hide();
    ui.searchBox->setTextWidget(ui.helpView);

    q->connect(ui.helpView, SIGNAL(sourceChanged(QUrl)), SLOT(viewSourceChanged(QUrl)));

    ui.toolbarAddress->hide();
    addressEntry = new QComboBox(ui.toolbarAddress);
    addressEntry->setEditable(true);
    addressEntry->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    ui.toolbarAddress->insertWidget(ui.actionGo, addressEntry);
    pub->connect(addressEntry, SIGNAL(activated(QString)), SLOT(openAddress(QString)));

    ui.menubar->insertMenu(ui.menuHelp->menuAction(), q->windowMenu());

    ui.bookmarkDock->hide();
    QAction *action = ui.bookmarkDock->toggleViewAction();
    action->setShortcut(QKeySequence(HelpBrowser::tr("Alt+B")));
    ui.menuView->addAction(action);

    ui.menuView->addAction(ui.toolbar->toggleViewAction());
    ui.menuView->addAction(ui.toolbarAddress->toggleViewAction());

    ui.actionGoForward->setShortcuts(QKeySequence::Forward);
    ui.actionGoBackward->setShortcuts(QKeySequence::Back);
    ui.actionFind->setShortcuts(QKeySequence::Find);
    ui.actionFindNext->setShortcuts(QKeySequence::FindNext);
    ui.actionFindPrevious->setShortcuts(QKeySequence::FindPrevious);
    ui.actionZoomIn->setShortcuts(QKeySequence::ZoomIn);
    ui.actionZoomOut->setShortcuts(QKeySequence::ZoomOut);
    ui.actionCopy->setShortcuts(QKeySequence::Copy);
    ui.actionPrint->setShortcuts(QKeySequence::Print);
    ui.actionClose->setShortcuts(QKeySequence::Close);

    q->initUi();
}

HelpView *HelpBrowser::view()
{
    return impl->ui.helpView;
}

void HelpBrowser::on_actionAddBookmark_triggered()
{
    QString label = impl->ui.helpView->documentTitle();
    QUrl url = impl->ui.helpView->source();

    if (!url.isValid())
        return;

    label = runBookmarkDialog(label, this);
    if (!label.isEmpty())
        impl->ui.bookmarkList->addBookmark(label, url);
}

void HelpBrowser::openAddress(const QString &text)
{
    if (text.isEmpty())
        return;
    QUrl url(text, QUrl::TolerantMode);
    if (!url.isValid())
    {
        moo::dlg::information(this, QString(), tr("Invalid URL"), QString(url.errorString()));
        return;
    }
    view()->setSource(url);
}

void HelpBrowser::viewSourceChanged(const QUrl &url)
{
    impl->addressEntry->setEditText(url.toString());
}

void HelpBrowser::on_actionGo_triggered()
{
    openAddress(impl->addressEntry->currentText());
}


void HelpBrowser::applyPrefs()
{
    view()->applyPrefs();
    moo::AppWindow::applyPrefs();
}


static void zoom(QWidget *w, double factor)
{
    QFont font = w->font();
    if (font.pointSize() > 0)
        font.setPointSize(qBound(6., factor * font.pointSize(), 10000.));
    else if (font.pixelSize() > 0)
        font.setPixelSize(qBound(6., factor * font.pixelSize(), 10000.));
    else
        qCritical("%s: oops", Q_FUNC_INFO);

    w->setFont(font);
}

void HelpBrowser::on_actionZoomIn_triggered()
{
    zoom(view(), 1.2);
}

void HelpBrowser::on_actionZoomOut_triggered()
{
    zoom(view(), 1./1.2);
}

void HelpBrowser::on_actionZoomNormal_triggered()
{
    view()->setFontFromPrefs();
}
