#include "moo/macros.h"
#include "moo/app.h"
#include "moo/singleapp.h"
#include "moo/appwindow.h"
#include "moo/utils.h"
#include "moo/debug.h"
#include <QtGui>
#ifdef Q_OS_MAC
#include <CoreFoundation/CoreFoundation.h>
#endif

using namespace moo;

namespace {

struct AppData {
    QString tempDir;
    QStringList tempFiles;
    QSettings *settings[2];

    bool inQuit;

    AppData() :
        inQuit(false)
    {
        settings[0] = settings[1] = 0;
    }
};

static AppData appData;

}

App::App(int &argc, char **argv,
         const QString &orgDomain,
         const QString &appName,
         const QString &appVersion) :
    QApplication(argc, argv)
{
    setOrganizationName(orgDomain);
    setOrganizationDomain(orgDomain);
    setApplicationName(appName);

#if QT_VERSION >= 0x040400
    setApplicationVersion(appVersion);
#endif

#if defined(Q_OS_MAC)
    setAttribute(Qt::AA_DontShowIconsInMenus);
#endif

#ifdef Q_OS_WIN32
    appData.settings[SettingsPrefs] = new QSettings(QSettings::IniFormat, QSettings::UserScope,
                                                    orgDomain, appName, this);
    appData.settings[SettingsState] = new QSettings(QSettings::IniFormat, QSettings::UserScope,
                                                    orgDomain, appName + "-state", this);
#else
    appData.settings[SettingsPrefs] = new QSettings(orgDomain, appName, this);
    appData.settings[SettingsState] = new QSettings(orgDomain, appName + "-state", this);
#endif

    mDebug("app") << "SettingsPrefs" << appData.settings[SettingsPrefs]->fileName();
    mDebug("app") << "SettingsState" << appData.settings[SettingsState]->fileName();

    setQuitOnLastWindowClosed(true);
}


QSettings *App::settings(SettingsType type)
{
    m_return_val_if_fail(type >= 0 && type < 2, new QSettings);
    return appData.settings[type];
}

static void deleteFile(const QString &name)
{
    QFileInfo fi(name);
    if (fi.isDir() && !fi.isSymLink())
    {
        QDir dir(name);
        QStringList files = dir.entryList();
        foreach (const QString &f, files)
            if (f != "." && f != "..")
                deleteFile(dir.filePath(f));
        QDir(fi.path()).rmdir(fi.fileName());
    }
    else
        QFile::remove(name);
}

App::~App()
{
    if (!appData.tempDir.isEmpty())
        deleteFile(appData.tempDir);
    foreach (const QString &name, appData.tempFiles)
        deleteFile(name);
}

static bool isRunning(App *app)
{
    bool running = false;
    QWidgetList list = app->topLevelWidgets();
    for (int i = 0; !running && i < list.size(); ++i)
    {
        QWidget *w = list.at(i);
        if (dynamic_cast<AppWindow*>(w))
            running = true;
    }
    return running;
}

void App::maybeQuit(QCloseEvent *e)
{
    if (e)
        e->ignore();

    if (appData.inQuit)
        return;

    appData.inQuit = true;
    closeAllWindows();
    appData.inQuit = false;

    if (e && !isRunning(this))
        e->accept();
}

void App::maybeQuit()
{
    maybeQuit(0);
}

bool App::event(QEvent *e)
{
    switch (e->type())
    {
        case QEvent::FileOpen:
            loadFile(static_cast<QFileOpenEvent *>(e)->file());
            return true;

        case QEvent::Close:
            maybeQuit(static_cast<QCloseEvent*>(e));
            return true;

        default:
            return QApplication::event(e);
    }
}


QString App::getUserDataDir(const QString &name)
{
#if QT_VERSION >= 0x040400
    QString dir = QDesktopServices::storageLocation(QDesktopServices::DataLocation);
#elif defined(Q_OS_UNIX)
    QString xdgDataHome = qgetenv("XDG_DATA_HOME");
    if (xdgDataHome.isEmpty())
        xdgDataHome = QDir::homePath() + "/.local/share";
    xdgDataHome += QString("/data/") + organizationName() + "/" + applicationName();
    QString dir = xdgDataHome;
#else
#error "Implement me"
#endif
    if (name.isEmpty())
        return dir;
    else
        return dir + "/" + name;
}

QString App::getUserDataFile(const QString &name)
{
    return getUserDataDir(name);
}

bool App::makeUserDataDir(const QString &name)
{
    QString path = getUserDataDir(name);
    bool ret = util::mkpath(path);
    if (!ret)
        qCritical() << "could not create directory '" << path << "'";
    return ret;
}

QString App::topTempDir()
{
    static QMutex mutex;

    QMutexLocker lock(&mutex);

    if (appData.tempDir.isEmpty())
    {
        QDir td = QDir::temp();
        qsrand(QDateTime::currentDateTime().toTime_t());

        for (int i = 0; i < 1000; ++i)
        {
            QString name;
            name.sprintf("-%08x", (quint32) qrand());
            name = applicationName() + name;
            if (td.mkdir(name))
            {
                appData.tempDir = td.filePath(name);
                mDebug("app") << "created temporary directory" << appData.tempDir;
                break;
            }
        }

        if (appData.tempDir.isEmpty())
        {
            // ??? now what?
            qCritical("%s: oops", Q_FUNC_INFO);
        }
    }

    return appData.tempDir;
}

QString App::fileNameInTempDir(const QString &name)
{
    return name.isEmpty() ? topTempDir() : topTempDir() + "/" + name;
}

QString App::tempFileName(const QString &name)
{
    static QMutex mutex;
    QMutexLocker lock(&mutex);

    static QMap<QString, int> counter;
    QString fullname;

    if (name.isEmpty())
    {
        fullname = QString::number(++counter[""]);
    }
    else if (counter.contains(name))
    {
        fullname = QString("%1-%2").arg(++counter[name]).arg(name);
    }
    else
    {
        fullname = name;
        counter[name] = 0;
    }

    return fileNameInTempDir(fullname);
}

QDir App::makeTempDir()
{
    QDir d(tempFileName());
    d.mkpath(".");
    return d;
}

QString App::saveTempFile(const QString &contents, QString &error, const QString &name)
{
    QFile file;
    if (!openTempFile(file, error, name))
        return QString();
    QByteArray bytes = contents.toUtf8();
    if (file.write(bytes) != bytes.size() || !file.flush())
    {
        error = file.errorString();
        file.remove();
        return QString();
    }
    file.close();
    return file.fileName();
}

bool App::openTempFile(QFile &file, QString &error, const QString &name)
{
    m_return_val_if_fail(!file.isOpen(), false);
    QString filename = tempFileName(name);
    file.setFileName(filename);
    if (!file.open(QIODevice::WriteOnly))
    {
        error = file.errorString();
        return false;
    }
    return true;
}


/////////////////////////////////////////////////////////////////////////////
// SingleApp
//

#define SINGLE_APP_MAGIC "SA01"
#define SINGLE_APP_ACK   "SAack"
#define SINGLE_APP_PING  "*ping*"

static QString makeSocketFilename(const QString &appId)
{
    QString userName = moo::util::userName().replace(QRegExp("[^A-Za-z0-9_.-]"), "_");

#ifdef Q_WS_X11
    // XXX add display
    return QString("%1/sa-%2-%3").arg(QDir::tempPath()).arg(appId).arg(userName);
#elif !defined(Q_OS_WIN32)
    return QString("%1/sa-%2-%3").arg(QDir::tempPath()).arg(appId).arg(userName);
#else
    return QString("sa-%1-%2").arg(appId).arg(userName);
#endif
}

static qint64 sendMessage(const QString &filename, const QByteArray &message, int timeout, QString &error)
{
    QLocalSocket socket;
    socket.connectToServer(filename);

    if (!socket.waitForConnected(timeout))
    {
        error = QString("error in QLocalSocket::waitForConnected: %1")
                   .arg(socket.errorString());
        return 0;
    }

    socket.write(SINGLE_APP_MAGIC, strlen(SINGLE_APP_MAGIC));
    socket.write(message.data());
    char zero = 0;
    socket.write(&zero, 1);

    if (!socket.waitForBytesWritten(timeout))
    {
        error = QString("error in QLocalSocket::waitForBytesWritten: %1")
                   .arg(socket.errorString());
        return 0;
    }

    if (!socket.waitForReadyRead(timeout))
    {
        error = QString("error in QLocalSocket::waitForReadyRead: %1")
                   .arg(socket.errorString());
        return 0;
    }

    QByteArray response = socket.readAll();
    if (!response.startsWith(SINGLE_APP_ACK) || response.size() < (int) strlen(SINGLE_APP_ACK) + 2)
    {
        error = "invalid response";
        return 0;
    }

    bool ok = false;
    qint64 pid = response.mid(strlen(SINGLE_APP_ACK), response.size() - strlen(SINGLE_APP_ACK) - 1).toLongLong(&ok);
    if (!ok)
    {
        error = "invalid response";
        return 0;
    }

    return pid;
}

static qint64 checkRunning(const QString &filename,
                           QLocalServer* &serverp)
{
    QString error;
    qint64 pid;
    if ((pid = sendMessage(filename, SINGLE_APP_PING, 3000, error)))
        return pid;

    QLocalServer *server = new QLocalServer;
    server->listen(filename);

#ifndef Q_OS_WIN32
    if (!server->isListening() && server->serverError() == QAbstractSocket::AddressInUseError)
    {
        qDebug() << filename;
        QFile::remove(filename);
        server->listen(filename);
    }
#endif

    if (server->isListening())
        serverp = server;
    else
        delete server;

    return 0;
}

SingleApp::SingleApp(const QString &id, int &/*argc*/, char **/*argv*/) :
    appId(id), filename(),
    disabled(false),
    server(0), pid(0)
{
    // parseArgs(argc, argv, appId, disabled);

    filename = makeSocketFilename(appId);
    if (!disabled)
        pid = checkRunning(filename, server);
    if (server)
        connect(server, SIGNAL(newConnection()), SLOT(newConnection()));
}

SingleApp::~SingleApp()
{
    if (server)
        server->close();
    delete server;
}

bool SingleApp::sendMessage(const QByteArray &message, int timeout)
{
    m_return_val_if_fail(server == 0, false);
    QString error;
    if (!::sendMessage(filename, message, timeout, error))
    {
        qCritical("%s", qPrintable(error));
        return false;
    }
    return true;
}

long SingleApp::processId() const
{
    return pid;
}

bool SingleApp::isRunning() const
{
    return pid != 0;
}

bool SingleApp::isDisabled() const
{
    return disabled;
}

void SingleApp::newConnection()
{
    m_return_if_fail(server != 0);
    QLocalSocket *socket = server->nextPendingConnection();
    m_return_if_fail(socket != 0);

    QByteArray buffer;

    while (true)
    {
        if (!socket->waitForReadyRead(3000))
        {
            qCritical("unterminated data from a socket");
            delete socket;
            return;
        }

        buffer += socket->readAll();
        if (!buffer.isEmpty() && buffer[buffer.size() - 1] == '\0')
            break;
    }

    if (buffer.mid(0, strlen(SINGLE_APP_MAGIC)) != SINGLE_APP_MAGIC)
    {
        qCritical("got invalid data from a socket");
        delete socket;
        return;
    }

    buffer = buffer.mid(strlen(SINGLE_APP_MAGIC), buffer.size() - strlen(SINGLE_APP_MAGIC) - 1);

    socket->write(SINGLE_APP_ACK, strlen(SINGLE_APP_ACK));
    socket->write(QByteArray::number(QCoreApplication::applicationPid()));
    char zero = 0;
    socket->write(&zero, 1);
    socket->waitForBytesWritten(500);
    delete socket;

    if (buffer == SINGLE_APP_PING)
        mDebug("singleapp") << "got ping";
    else
        emit messageReceived(buffer);
}
