#include "docwindow-p.h"
#include "dialogs.h"
#include "macros.h"
#include "prefs.h"
#include "utils.h"
#include "docmanager.h"

using namespace moo;

QList<DocumentWindow*> DocumentWindowPrivate::windows;

DocumentWindow::DocumentWindow(const QString &name) :
    AppWindow(name),
    impl(this)
{
    // fake filename, for setWindowModified to work
    setWindowFilePath("Untitled");
}

DocumentWindow::~DocumentWindow()
{
}

void DocumentWindow::initUi()
{
    connect(document(), SIGNAL(docStateChanged()), SLOT(docStateChanged()));
    connect(document(), SIGNAL(modificationChanged(bool)), SLOT(setWindowModified(bool)));
    connect(document(), SIGNAL(urlChanged(QUrl)), SLOT(setWindowFileUrl(QUrl)));

    connectAction("actionOpen", SLOT(actionOpen()));
    connectAction("actionSave", SLOT(actionSave()));
    connectAction("actionSaveAs", SLOT(actionSaveAs()));

    setActionShortcuts("actionOpen", QKeySequence::Open);
    setActionShortcuts("actionSave", QKeySequence::Save);
    setActionShortcuts("actionReload", QKeySequence::Refresh);

    AppWindow::initUi();
}

void DocumentWindow::present(DocumentWindow *old)
{
    Q_UNUSED(old);

#ifdef Q_OS_MAC
    if (old)
        move(old->x() + 22, old->y() + 22); // FIXME
#endif

    if (document()->isUntitled())
        setWindowFilePath(document()->displayBasename());

    show();
}

void DocumentWindow::docStateChanged()
{
    switch (document()->documentState())
    {
        case Document::Saving:
            showMessage(tr("Saving"));
            break;
        case Document::Loading:
            showMessage(tr("Loading"));
            break;
        case Document::Idle:
            clearMessage();
            break;
    }
}

void DocumentWindow::saveFile(const QUrl &url, const QString &type, bool close_on_save, bool ask_filename)
{
    m_return_if_fail(document()->documentState() == Document::Idle);

    QUrl realUrl = url;
    QString useType = type;
    if (realUrl.isEmpty())
        realUrl = document()->url();
    if (ask_filename || realUrl.isEmpty())
        realUrl = saveFileDialog(&useType);
    if (realUrl.isEmpty())
        return;

    QString error;
    bool success = document()->save(realUrl, useType, error);

    if (!success)
        moo::dlg::saveError(this, moo::util::uriDisplayName(url), error);
    else if (close_on_save)
        QTimer::singleShot(0, this, SLOT(close()));
}

void DocumentWindow::loadFile(const QUrl &url, DocumentWindow *active)
{
    m_return_if_fail(!url.isEmpty());

    if (DocumentWindow *old = DocumentWindowPrivate::find(url))
    {
        RecentMenu::addUrl(old->document()->url());
        old->raise();
        old->activateWindow();
        return;
    }

    if (active && active->document()->isEmpty())
    {
        active->loadFile(url);
    }
    else
    {
        DocumentWindow *win = DocumentManager::createWindow();

        if (!win->loadFile(url))
        {
            delete win;
            return;
        }

        win->present(active);
    }
}

void DocumentWindow::openFile(const QUrl &url, DocumentWindow *window)
{
    if (!window)
        window = dynamic_cast<DocumentWindow*>(QApplication::activeWindow());

    if (!window)
    {
        QWidgetList windows = QApplication::topLevelWidgets();
        foreach (QWidget *w, windows)
            if ((window = dynamic_cast<DocumentWindow*>(w)))
                break;
    }

    loadFile(url, window);
}

bool DocumentWindow::loadFile(const QUrl &url)
{
    QString error;
    bool result = document()->load(url, error);
    if (!result)
        moo::dlg::openError(this, moo::util::uriDisplayName(url), error);
    return result;
}

void DocumentWindow::actionOpen()
{
    QList<QUrl> urls = openFileDialog();
    if (!urls.isEmpty())
        foreach (const QUrl &url, urls)
            loadFile(url, this);
}

void DocumentWindow::actionSave()
{
    saveFile(QString(), QString(), false);
}

void DocumentWindow::actionSaveAs()
{
    saveFile(QString(), QString(), false, true);
}


static void demandAttention(DocumentWindow *window)
{
//     if (!qApp->activeWindow())
//     {
//         window->raise();
//         qApp->alert(window);
//     }
//     else
//     {
        window->raise();
        window->activateWindow();
//     }
}

void DocumentWindow::closeEvent(QCloseEvent *e)
{
    if (document()->documentState() != Document::Idle)
    {
        demandAttention(this);
//         raise();
//         activateWindow();
        QApplication::beep();
        e->ignore();
        return;
    }

//     if (document()->isModified() && (ws()->mode() == Worksheet::File || !ws()->isUntitled()))
    if (document()->needSave())
    {
        demandAttention(this);
//         raise();
//         activateWindow();

        QMessageBox *msgBox = moo::dlg::saveChanges(this);

        switch (msgBox->exec())
        {
            case QMessageBox::Cancel:
                e->ignore();
                return;
            case QMessageBox::Save:
                saveFile(QString(), QString(), true);
                e->ignore();
                return;
        }
    }

    AppWindow::closeEvent(e);
}

QMenu *DocumentWindow::recentMenu()
{
    return new RecentMenu(this, this);
}


/**************************************************************************/
/* RecentMenu
 */

#define MAX_RECENT_FILES 10
#define PREFS_RECENT_FILES "files/recent-files"

struct moo::RecentMenuPrivate {
    QList<QAction*> actions;
    bool update_timer_started;
    DocumentWindow *window;
    static QList<RecentMenu*> menus;

    RecentMenuPrivate(RecentMenu *pub, DocumentWindow *window);
    ~RecentMenuPrivate();
};

QList<RecentMenu*> RecentMenuPrivate::menus;

RecentMenuPrivate::RecentMenuPrivate(RecentMenu *menu, DocumentWindow *window) :
    update_timer_started(false), window(window)
{
    for (int i = 0; i < MAX_RECENT_FILES; ++i)
    {
        QAction *a = new QAction(menu);
        QObject::connect(a, SIGNAL(triggered()), menu, SLOT(actionTriggered()));
        actions.append(a);
    }
}

RecentMenuPrivate::~RecentMenuPrivate()
{
}

RecentMenu::RecentMenu(DocumentWindow *window, QWidget *parent) :
    QMenu(parent), impl(new RecentMenuPrivate(this, window))
{
    setTitle(tr("Open &Recent"));

    foreach (QAction *a, impl->actions)
        addAction(a);

    addSeparator();
    QAction *clear_menu = new QAction(tr("Clear Menu"), this);
    connect(clear_menu, SIGNAL(triggered()), SLOT(clearMenu()));
    addAction(clear_menu);

    RecentMenuPrivate::menus.append(this);
    updateMenu();
}

RecentMenu::~RecentMenu()
{
    RecentMenuPrivate::menus.removeAll(this);
    delete impl;
}

void RecentMenu::queueUpdate()
{
    if (!impl->update_timer_started)
    {
        impl->update_timer_started = true;
        QTimer::singleShot(0, this, SLOT(updateMenu()));
    }
}

void RecentMenu::updateMenu()
{
    impl->update_timer_started = false;

    QStringList files = stateValue(PREFS_RECENT_FILES).toStringList();
    int n = qMin(files.size(), MAX_RECENT_FILES);

    for (int i = 0; i < n; ++i)
    {
        QString text = tr("&%1. %2").arg(i + 1).arg(QFileInfo(files[i]).fileName());
        impl->actions.at(i)->setText(text);
        impl->actions.at(i)->setStatusTip(util::uriDisplayName(files[i]));
        impl->actions.at(i)->setData(files[i]);
        impl->actions.at(i)->setVisible(true);
    }

    for (int i = n; i < MAX_RECENT_FILES; ++i)
        impl->actions.at(i)->setVisible(false);

     setEnabled(n > 0);
}

void RecentMenu::actionTriggered()
{
    QAction *action = dynamic_cast<QAction*>(sender());
    if (action)
    {
        QString filename = action->data().toString();
        if (!filename.isEmpty())
            DocumentWindow::openFile(filename, impl->window);
    }
}

void RecentMenu::clearMenu()
{
    clearList();
}

void RecentMenu::queueUpdateAll()
{
    foreach (RecentMenu *m, RecentMenuPrivate::menus)
        m->queueUpdate();
}

void RecentMenu::addUrl(const QUrl &url)
{
    QStringList files = stateValue(PREFS_RECENT_FILES).toStringList();
    files.removeAll(url.toString());
    files.prepend(url.toString());
    while (files.size() > MAX_RECENT_FILES)
        files.removeLast();
    setStateValue(PREFS_RECENT_FILES, files);
    queueUpdateAll();
}

void RecentMenu::clearList()
{
    unsetStateValue(PREFS_RECENT_FILES);
    queueUpdateAll();
}
